# --- Day 2: Bathroom Security ---

# You arrive at Easter Bunny Headquarters under cover of darkness. However, you
# left in such a rush that you forgot to use the bathroom! Fancy office
# buildings like this one usually have keypad locks on their bathrooms, so you
# search the front desk for the code.

# "In order to improve security," the document you find says, "bathroom codes
# will no longer be written down. Instead, please memorize and follow the
# procedure below to access the bathrooms."

# The document goes on to explain that each button to be pressed can be found
# by starting on the previous button and moving to adjacent buttons on the
# keypad: U moves up, D moves down, L moves left, and R moves right. Each line
# of instructions corresponds to one button, starting at the previous button
# (or, for the first line, the "5" button); press whatever button you're on at
# the end of each line. If a move doesn't lead to a button, ignore it.

# You can't hold it much longer, so you decide to figure out the code as you
# walk to the bathroom. You picture a keypad like this:

# 1 2 3
# 4 5 6
# 7 8 9

# Suppose your instructions are:

# ULL
# RRDDD
# LURDL
# UUUUD

#     You start at "5" and move up (to "2"), left (to "1"), and left (you
# can't, and stay on "1"), so the first button is 1.
#     Starting from the previous button ("1"), you move right twice (to "3")
# and then down three times (stopping at "9" after two moves and ignoring the
# third), ending up with 9.
#     Continuing from "9", you move left, up, right, down, and left, ending
# with 8.
#     Finally, you move up four times (stopping at "2"), then down once, ending
# with 5.

# So, in this example, the bathroom code is 1985.

# Your puzzle input is the instructions from the document you found at the
# front desk. What is the bathroom code?

from typing import Optional, Tuple, Union

with open("files/P2.txt") as f:
    instructions = [line for line in f.read().strip().split()]


def move_right(position: Tuple[int, int]) -> Tuple[int, int]:
    _x = position[0] + 1
    if 0 <= _x <= 2:
        return _x, position[1]
    else:
        return position[0], position[1]


def move_down(position: Tuple[int, int]) -> Tuple[int, int]:
    _y = position[1] - 1
    if 0 <= _y <= 2:
        return position[0], _y
    else:
        return position[0], position[1]


def move_left(position: Tuple[int, int]) -> Tuple[int, int]:
    _x = position[0] - 1
    if 0 <= _x <= 2:
        return _x, position[1]
    else:
        return position[0], position[1]


def move_up(position: Tuple[int, int]) -> Tuple[int, int]:
    _y = position[1] + 1
    if 0 <= _y <= 2:
        return position[0], _y
    else:
        return position[0], position[1]


def pos2num(position: Tuple[int, int]) -> Optional[int]:
    num_dict = {
        1: (0, 2),
        2: (1, 2),
        3: (2, 2),
        4: (0, 1),
        5: (1, 1),
        6: (1, 2),
        7: (0, 0),
        8: (1, 0),
        9: (2, 0),
    }
    return next((k for k, v in num_dict.items() if v == position), None)


def part_1() -> None:
    secret = []
    for instruction in instructions:
        # 0,0 is the x,y coordinates of the bottom left position
        pos = (1, 1)
        for button in instruction:
            if button == "D":
                pos = move_down(pos)
            elif button == "U":
                pos = move_up(pos)
            elif button == "R":
                pos = move_right(pos)
            elif button == "L":
                pos = move_left(pos)
        secret.append(pos)
    secret_code = "".join(map(str, (pos2num(num) for num in secret)))
    print(f"The secret code is {secret_code}")


# --- Part Two ---

# You finally arrive at the bathroom (it's a several minute walk from the lobby
# so visitors can behold the many fancy conference rooms and water coolers on
# this floor) and go to punch in the code. Much to your bladder's dismay, the
# keypad is not at all like you imagined it. Instead, you are confronted with
# the result of hundreds of man-hours of bathroom-keypad-design meetings:

#     1
#   2 3 4
# 5 6 7 8 9
#   A B C
#     D

# You still start at "5" and stop when you're at an edge, but given the same
# instructions as above, the outcome is very different:

#     You start at "5" and don't move at all (up and left are both edges),
# ending at 5.
#     Continuing from "5", you move right twice and down three times (through
# "6", "7", "B", "D", "D"), ending at D.
#     Then, from "D", you move five more times (through "D", "B", "C", "C",
# "B"), ending at B.
#     Finally, after five more moves, you end at 3.

# So, given the actual keypad layout, the code would be 5DB3.

# Using the same instructions in your puzzle input, what is the correct
# bathroom code?


def move_right_(position: Union[int, str]) -> Union[int, str]:
    num_dict: dict[Union[int, str], Union[int, str]] = {
        1: 1,
        2: 3,
        3: 4,
        4: 4,
        5: 6,
        6: 7,
        7: 8,
        8: 9,
        9: 9,
        "A": "B",
        "B": "C",
        "C": "C",
        "D": "D",
    }
    return num_dict[position]


def move_down_(position: Union[int, str]) -> Union[int, str]:
    num_dict: dict[Union[int, str], Union[int, str]] = {
        1: 3,
        2: 6,
        3: 7,
        4: 8,
        5: 5,
        6: "A",
        7: "B",
        8: "C",
        9: 9,
        "A": "A",
        "B": "D",
        "C": "C",
        "D": "D",
    }
    return num_dict[position]


def move_left_(position: Union[int, str]) -> Union[int, str]:
    num_dict: dict[Union[int, str], Union[int, str]] = {
        1: 1,
        2: 2,
        3: 2,
        4: 3,
        5: 5,
        6: 5,
        7: 6,
        8: 7,
        9: 8,
        "A": "A",
        "B": "A",
        "C": "B",
        "D": "D",
    }
    return num_dict[position]


def move_up_(position: Union[int, str]) -> Union[int, str]:
    num_dict: dict[Union[int, str], Union[int, str]] = {
        1: 1,
        2: 2,
        3: 1,
        4: 4,
        5: 5,
        6: 2,
        7: 3,
        8: 4,
        9: 9,
        "A": 6,
        "B": 7,
        "C": 8,
        "D": "B",
    }
    return num_dict[position]


def part_2() -> None:
    secret = []
    for instruction in instructions:
        pos: Union[int, str] = 5
        for button in instruction:
            if button == "D":
                pos = move_down_(pos)
            elif button == "U":
                pos = move_up_(pos)
            elif button == "R":
                pos = move_right_(pos)
            elif button == "L":
                pos = move_left_(pos)
        secret.append(pos)

    secret_code = "".join(map(str, (char for char in secret)))
    print(f"The bathroom code is {secret_code}")


if __name__ == "__main__":
    part_1()
    part_2()
